# TODO NOT IMPLEMENTED YET
# TODO dynamic setters which return self. e.g.: indexDef.name('test_name').is_pk().is_dense()
# TODO check types for each attrs of index definition
# TODO check possible values for attrs field_type, index_type, collate_mode


class IndexDefinition(dict):
    """IndexDefinition is a dictionary subclass which allows to construct and manage indexes more efficiently.
    NOT IMPLEMENTED YET. USE FIELDS DESCRIPTION ONLY.

    #### Arguments:
        name (str): An index name.
        json_paths (:obj:`list` of :obj:`str`): A name for mapping a value to a json field.
        field_type (str): A type of field. Possible values are: `int`, `int64`, `double`, `string`, `bool`,
        `composite`, `float_vector`.
        index_type (str): An index type. Possible values are: `hash`, `tree`, `text`, `-`, `hnsw`, `vec_bf`, `ivf`.
        is_pk (bool): True if a field is a primary key.
        is_array (bool): True if an index is an array.
        is_dense (bool): True if an index is dense. Reduce the index size. Saves 8 bytes per unique key value for 'hash'
            and 'tree' index types. For '-' index type saves 4-8 bytes per each element. Useful for indexes with
            high selectivity, but for tree and hash indexes with low selectivity can seriously decrease update
            performance.
        is_no_column (bool): True if allows to disable column subindex. Reduces the index size.
            Allows to save ~(`stored_type_size` * `namespace_items_count`) bytes, where `stored_type_size` is the size
            of the type stored in the index, and `namespace_items_count` is the number of items in the namespace.
            May reduce performance.
        is_sparse (bool): True if a value of an index may be not presented.
        collate_mode (str): Sets an order of values by collate mode. Possible values are:
            `none`, `ascii`, `utf8`, `numeric`, `custom`.
        sort_order_letters (str): Order for a sort sequence for a custom collate mode.
        config (dict): A config for a fulltext and float_vector engine.
        [More about `fulltext`](https://github.com/Restream/reindexer/blob/master/fulltext.md) or
        [More about `float_vector`](https://github.com/Restream/reindexer/blob/master/float_vector.md).
    """

    def __getitem__(self, attr):
        self._raise_if_key_error(attr)
        return super(IndexDefinition, self).get(attr)

    def __setitem__(self, attr, value):
        self._raise_if_key_error(attr)
        super(IndexDefinition, self).update({attr: value})
        return self

    def update(self, *args, **kwargs):
        raise NotImplementedError(
            'Bulk update is not implemented for IndexDefinition instance')

    @staticmethod
    def _get_known_attrs() -> list[str]:
        return ['name', 'json_paths', 'field_type', 'index_type', 'is_pk',
                'is_array', 'is_dense', 'is_sparse', 'collate_mode', 'sort_order_letters', 'expire_after', 'config']

    def _raise_if_key_error(self, attr):
        known_attrs = self._get_known_attrs()
        if attr not in known_attrs:
            raise KeyError("Invalid key '{0}'. Known keys are: '{1}'".format(
                attr, ', '.join(known_attrs)))



import os
import platform
import shutil
from datetime import timedelta
from pathlib import Path

import pytest
from hamcrest import *

from pyreindexer.exceptions import ApiError
from pyreindexer.query import CondType
from tests.helpers.api import ConnectorApi
from tests.helpers.base_helper import get_ns_description, get_ns_items
from tests.helpers.server_helper import ReindexerServer
from tests.test_data.auth import users_yml
from tests.test_data.constants import index_definition, item_definition


CONNECTIONS = []


@pytest.fixture(scope="module", autouse=True)
def rx_server(request):
    if request.config.getoption("--mode") == "builtin":
        pytest.skip("auth tests run only for cproto mode")
    else:
        rx_bin_path = request.config.getoption("--rx_bin_path")
        storage = f"/tmp/reindex_test_auth"
        if os.path.exists(storage):
            shutil.rmtree(storage, ignore_errors=True)
        Path(storage).mkdir(parents=True)
        with open(os.path.join(storage, "users.yml"), "w") as file:
            file.write(users_yml)

        server = ReindexerServer(rx_bin_path=rx_bin_path, http_port=9089, rpc_port=6535, storage=storage, auth=True,
                                 user="owner", password="owner")
        server.run(module=request.node.nodeid)
        yield
        server.terminate()


@pytest.fixture(scope="module")
def db(request):
    db = ConnectorApi("cproto://owner:owner@127.0.0.1:6535/test_db")
    yield db
    db.close()
    shutil.rmtree("tmp/", ignore_errors=True)


@pytest.fixture
def db_auth(request):
    def _inner(auth):
        db = ConnectorApi(f"cproto://{auth[0]}:{auth[1]}@127.0.0.1:6535/test_db")
        CONNECTIONS.append(db)
        return db

    yield _inner


@pytest.fixture(autouse=True)
def close_connections(db, db_auth):
    yield
    for db_conn in CONNECTIONS:
        db_conn.close()
    for ns in db.namespace.enumerate():
        if not ns["name"].startswith("#"):
            db.namespace.drop(ns["name"])
    CONNECTIONS.clear()


class TestAuth:

    @pytest.mark.skipif(platform.system().lower() == "darwin", reason="doesn't work on Mac OS for now")
    def test_cant_connect_to_db_with_invalid_pass(self, db_auth):
        # When ("Connect to DB")
        # Then ("Connection is not established")
        auth = ("invalid_pass", "invalid_pass")
        err_msg = f"Errors occurred when parsing the URL to mask user credentials"
        assert_that(calling(db_auth).with_args(auth),
                    raises(ApiError, pattern=err_msg))

    @pytest.mark.parametrize("auth", [
        ("owner", "owner"),
        ("db_admin", "dbadmin"),
        ("data_write", "datawrite"),
        ("data_read", "dataread")
    ])
    def test_auth_db_admin_can_open_namespace(self, db_auth, auth):
        # When ("Open namespace")
        # Then ("Namespace is opened")
        ns_name = "new_ns_auth"
        db_auth = db_auth(auth)
        db_auth.namespace.open(ns_name)

    @pytest.mark.parametrize("auth, err_msg", [
        (("error", "error"), "Unauthorized")
    ])
    def test_unauthorized_user_cant_open_namespace(self, db_auth, auth, err_msg):
        # When ("Open namespace")
        # Then ("Namespace is not opened")
        ns_name = "new_ns_auth"
        db_auth = db_auth(auth)
        assert_that(calling(db_auth.namespace.open).with_args(ns_name),
                    raises(ApiError, pattern=err_msg))

    @pytest.mark.parametrize("auth", [
        ("owner", "owner"),
        ("db_admin", "dbadmin"),
        ("data_write", "datawrite"),
        ("data_read", "dataread")
    ])
    def test_all_auth_users_can_get_namespaces_list(self, db, namespace, db_auth, auth):
        # When ("Get namespaces list")
        ns_name = "new_ns_auth"
        db_auth = db_auth(auth)
        namespaces_list = db_auth.namespace.enumerate()
        # Then ("Namespaces list is returned")
        assert_that(namespaces_list, has_item(has_entries(name=namespace)))

    @pytest.mark.parametrize("auth", [
        ("owner", "owner"),
        ("db_admin", "dbadmin"),
        ("data_write", "datawrite"),
        ("data_read", "dataread")
    ])
    def test_all_auth_users_can_sql_select(self, db, db_auth, namespace, index, item, auth):
        # Given("Create namespace with id and item")
        # When ("Execute SQL query SELECT")
        db_auth = db_auth(auth)
        query = f"SELECT * FROM {namespace}"
        items_list = list(db_auth.query.sql(query))
        # Then ("Check that selected item is in result")
        assert_that(items_list, equal_to([item]))

    @pytest.mark.parametrize("auth", [
        ("data_write", "datawrite"),
        ("data_read", "dataread")
    ])
    def test_auth_data_write_cant_create_index(self, db, db_auth, namespace, auth):
        # Given("Create namespace")
        # When ("Create index")
        # Then ("Index is not created")
        db_auth = db_auth(auth)
        role = auth[0]
        err_msg = f"Forbidden: need role db_admin of db 'test_db' user 'da...{role[-2:]}' have role={role}"
        assert_that(calling(db_auth.index.create).with_args(namespace, index_definition),
                    raises(ApiError, pattern=err_msg))

    @pytest.mark.parametrize("auth", [
        ("owner", "owner"),
        ("db_admin", "dbadmin")
    ])
    def test_auth_db_admin_can_create_index(self, db, db_auth, namespace, auth):
        # Given("Create namespace")
        # When ("Create index")
        db_auth = db_auth(auth)
        db_auth.index.create(namespace, index_definition)
        # Then ("Index is created")
        ns_entry = get_ns_description(db, namespace)
        assert_that(ns_entry, has_item(has_entry("indexes", has_item(index_definition))))

    @pytest.mark.parametrize("auth", [
        ("owner", "owner"),
        ("db_admin", "dbadmin"),
        ("data_write", "datawrite"),
        ("data_read", "dataread")
    ])
    def test_all_auth_users_can_query_select_where(self, db, db_auth, namespace, index, items, auth):
        # Given("Create namespace with index and items")
        # When ("Make select query")
        db_auth = db_auth(auth)
        query = db_auth.query.new(namespace)
        query_result = list(query.where("id", CondType.CondEq, 3).execute(timeout=timedelta(seconds=1)))
        # Then ("Check that selected item is in result")
        assert_that(query_result, equal_to([items[3]]))

    @pytest.mark.parametrize("auth", [
        ("owner", "owner"),
        ("db_admin", "dbadmin"),
        ("data_write", "datawrite")
    ])
    def test_auth_data_write_can_tx_item_insert(self, db, db_auth, namespace, index, auth):
        # Given("Create namespace with index")
        # When ("Insert item into namespace via tx")
        db_auth = db_auth(auth)
        transaction = db_auth.tx.begin(namespace)
        transaction.insert_item(item_definition)
        transaction.commit(timeout=timedelta(milliseconds=1000))
        # Then ("Item was added")
        select_result = get_ns_items(db_auth, namespace)
        assert_that(select_result, equal_to([item_definition]))

    def test_auth_data_read_cant_begin_tx(self, db, db_auth, namespace, index):
        # Given("Create namespace with index")
        # When ("Begin transaction")
        # then ("Transaction is not started")
        db_auth = db_auth(("data_read", "dataread"))
        err_msg = "Forbidden: need role data_write of db 'test_db' user 'da...ad' have role=data_read"
        assert_that(calling(db_auth.tx.begin).with_args(namespace), raises(ApiError, pattern=err_msg))



from hamcrest.core.base_matcher import BaseMatcher


class CloseToDict(BaseMatcher):
    def __init__(self, expected, delta):
        self.expected = expected
        self.delta = delta

    def _matches(self, actual):
        if not isinstance(actual, dict) or not isinstance(self.expected, dict):
            return False

        for key, expected_value in self.expected.items():
            if key not in actual:
                return False
            if not self.matches_values(expected_value, actual[key]):
                return False

        # Check if there are any redundant keys
        for key in actual:
            if key not in self.expected:
                return False

        return True

    def matches_values(self, expected_value, actual_value):
        if isinstance(expected_value, (int, float)):
            # If the value is a number, check with delta
            return expected_value - self.delta <= actual_value <= expected_value + self.delta
        elif isinstance(expected_value, list):
            # If the value is a list, check each element
            if not isinstance(actual_value, list) or len(actual_value) != len(expected_value):
                return False
            for i, expected_elem in enumerate(expected_value):
                if not self.matches_values(expected_elem, actual_value[i]):
                    return False
            return True
        elif isinstance(expected_value, dict):
            # If the value is a dictionary, compare entire dictionaries using the same method
            return CloseToDict(expected_value, self.delta)._matches(actual_value)
        else:
            # For other values check for strict equality
            return expected_value == actual_value

    def describe_to(self, description):
        description.append_text("a dictionary close to ").append_description_of(self.expected)


class CloseToItems(CloseToDict):
    def _matches(self, actual):
        if not isinstance(actual, list) or not isinstance(self.expected, list):
            return False

        if len(actual) != len(self.expected):
            return False

        for expected_dict, actual_dict in zip(self.expected, actual):
            if not CloseToDict(expected_dict, self.delta)._matches(actual_dict):
                return False

        return True

    def describe_to(self, description):
        description.append_text("a list of dictionaries close to ").append_description_of(self.expected)


def close_to_dict(expected, delta=0.01):
    """ Проверить, что 2 словаря приблизительно (с учетом дельты для числовых типов) равны друг другу
    :param expected: словарь для сравнения
    :param delta: дельта для приблизительного сравнения числовых типов
    """
    return CloseToDict(expected, delta)


def close_to_items(expected, delta=0.01):
    """ Проверить, что 2 списка словарей приблизительно (с учетом дельты для числовых типов) равны друг другу
    :param expected: список словарей для сравнения
    :param delta: дельта для приблизительного сравнения числовых типов
    """
    return CloseToItems(expected, delta)



import time
from datetime import timedelta

import pytest
from hamcrest import *

from pyreindexer.exceptions import TransactionError
from pyreindexer.query import CondType
from tests.helpers.base_helper import get_ns_items, get_ns_vect_items,random_vector
from tests.helpers.matchers import close_to_dict
from tests.helpers.transaction import *
from tests.test_data.constants import item_definition, vector_index_bf, vector_index_hnsw, vector_index_ivf


class TestCrudTransaction:
    def test_negative_commit_after_rollback(self, db, namespace, index):
        # Given("Create namespace")
        # When ("Start new transaction")
        transaction = db.tx.begin(namespace)
        # Then ("Rollback transaction")
        transaction.rollback()
        # Then ("Commit transaction")
        assert_that(calling(transaction.commit).with_args(),
                    raises(TransactionError, matching=has_string("Transaction is over")))

    def test_negative_rollback_after_commit(self, db, namespace, index):
        # Given("Create namespace")
        # When ("Start new transaction")
        transaction = db.tx.begin(namespace)
        # Then ("Commit transaction")
        transaction.commit()
        # Then ("Rollback transaction")
        assert_that(calling(transaction.rollback).with_args(),
                    raises(TransactionError, matching=has_string("Transaction is over")))

    def test_negative_insert_after_rollback(self, db, namespace, index):
        # Given("Create namespace with index")
        # When ("Start new transaction")
        transaction = db.tx.begin(namespace)
        # Then ("Rollback transaction")
        transaction.rollback()
        # Then ("Insert transaction")
        assert_that(calling(transaction.insert_item).with_args(item_definition),
                    raises(TransactionError, matching=has_string("Transaction is over")))

    def test_negative_update_after_rollback(self, db, namespace, index):
        # Given("Create namespace with index")
        # When ("Start new transaction")
        transaction = db.tx.begin(namespace)
        # Then ("Rollback transaction")
        transaction.rollback()
        # Then ("Update transaction")
        assert_that(calling(transaction.update_item).with_args(item_definition),
                    raises(TransactionError, matching=has_string("Transaction is over")))

    def test_negative_upsert_after_rollback(self, db, namespace, index):
        # Given("Create namespace with index")
        # When ("Start new transaction")
        transaction = db.tx.begin(namespace)
        # Then ("Rollback transaction")
        transaction.rollback()
        # Then ("Upsert transaction")
        assert_that(calling(transaction.upsert_item).with_args(item_definition),
                    raises(TransactionError, matching=has_string("Transaction is over")))

    def test_negative_delete_after_rollback(self, db, namespace, index):
        # Given("Create namespace with index")
        # When ("Start new transaction")
        transaction = db.tx.begin(namespace)
        # Then ("Rollback transaction")
        transaction.rollback()
        # Then ("Delete transaction")
        assert_that(calling(transaction.delete_item).with_args(item_definition),
                    raises(TransactionError, matching=has_string("Transaction is over")))

    def test_negative_insert_after_commit(self, db, namespace, index):
        # Given("Create namespace with index")
        # When ("Start new transaction")
        transaction = db.tx.begin(namespace)
        # Then ("Commit transaction")
        transaction.commit()
        # Then ("Insert transaction")
        assert_that(calling(transaction.insert_item).with_args(item_definition),
                    raises(TransactionError, matching=has_string("Transaction is over")))

    def test_negative_update_after_commit(self, db, namespace, index):
        # Given("Create namespace with index")
        # When ("Start new transaction")
        transaction = db.tx.begin(namespace)
        # Then ("Commit transaction")
        transaction.commit()
        # Then ("Update transaction")
        assert_that(calling(transaction.update_item).with_args(item_definition),
                    raises(TransactionError, matching=has_string("Transaction is over")))

    def test_negative_upsert_after_commit(self, db, namespace, index):
        # Given("Create namespace with index")
        # When ("Start new transaction")
        transaction = db.tx.begin(namespace)
        # Then ("Commit transaction")
        transaction.commit()
        # Then ("Upsert transaction")
        assert_that(calling(transaction.upsert_item).with_args(item_definition),
                    raises(TransactionError, matching=has_string("Transaction is over")))

    def test_negative_delete_after_commit(self, db, namespace, index):
        # Given("Create namespace with index")
        # When ("Start new transaction")
        transaction = db.tx.begin(namespace)
        # Then ("Commit transaction")
        transaction.commit()
        # Then ("Delete transaction")
        assert_that(calling(transaction.delete_item).with_args(item_definition),
                    raises(TransactionError, matching=has_string("Transaction is over")))

    def test_create_item_insert(self, db, namespace, index):
        # Given("Create namespace with index")
        # When ("Insert item into namespace")
        transaction = db.tx.begin(namespace, timeout=timedelta(milliseconds=1000))
        transaction.insert_item(item_definition)
        transaction.commit(timeout=timedelta(milliseconds=1000))
        # Then ("Check that item is added")
        select_result = get_ns_items(db, namespace)
        assert_that(select_result, has_length(1), "Transaction: item wasn't created")
        assert_that(select_result, has_item(item_definition), "Transaction: item wasn't created")

    @pytest.mark.parametrize("vector_index", [vector_index_bf, vector_index_hnsw, vector_index_ivf])
    def test_insert_item_with_vector_index(self, db, namespace, index, vector_index):
        # Given("Create namespace")
        # When ("Add index")
        db.index.create(namespace, vector_index)
        # When ("Insert item into namespace")
        transaction = db.tx.begin(namespace)
        dimension = vector_index["config"]["dimension"]
        item = {"id": 0, "vec": random_vector(dimension)}
        transaction.insert_item(item)
        transaction.commit()
        # Then ("Check that item is added")
        select_result = get_ns_vect_items(db, namespace)
        assert_that(select_result, has_length(1), "Item wasn't created")
        assert_that(select_result[0], close_to_dict(item))

    def test_create_item_insert_with_precepts(self, db, namespace, index):
        # Given("Create namespace with index")
        # When ("Insert items into namespace")
        transaction = db.tx.begin(namespace)
        number_items = 5
        for i in range(number_items):
            transaction.insert_item({'id': 100, 'field': f'value{100 + i}'}, ['id=serial()'])
        count = transaction.commit_with_count()
        assert_that(count, equal_to(number_items), "Transaction: items wasn't created")
        # Then ("Check that item is added")
        select_result = get_ns_items(db, namespace)
        assert_that(select_result, has_length(number_items), "Transaction: items wasn't created")
        for i in range(number_items):
            assert_that(select_result[i],
                        equal_to({'id': i + 1, 'field': f'value{100 + i}'}),
                        "Transaction: items wasn't created")

    def test_create_item_upsert(self, db, namespace, index):
        # Given("Create namespace with index")
        # When ("Upsert item into namespace")
        upsert_item_transaction(db, namespace, item_definition)
        # Then ("Check that item is added")
        select_result = get_ns_items(db, namespace)
        assert_that(select_result, has_length(1), "Transaction: item wasn't created")
        assert_that(select_result, has_item(item_definition), "Transaction: item wasn't created")

    def test_update_item_upsert(self, db, namespace, index, item):
        # Given("Create namespace with item")
        # When ("Upsert item")
        item_definition_updated = {'id': 100, 'val': "new_value"}
        upsert_item_transaction(db, namespace, item_definition_updated)
        # Then ("Check that item is updated")
        select_result = get_ns_items(db, namespace)
        assert_that(select_result, has_length(1), "Transaction: item wasn't updated")
        assert_that(select_result, has_item(item_definition_updated), "Transaction: item wasn't updated")

    def test_update_item_update(self, db, namespace, index, item):
        # Given("Create namespace with item")
        # When ("Update item")
        item_definition_updated = {'id': 100, 'val': "new_value"}
        update_item_transaction(db, namespace, item_definition_updated)
        # Then ("Check that item is updated")
        select_result = get_ns_items(db, namespace)
        assert_that(select_result, has_length(1), "Transaction: item wasn't updated")
        assert_that(select_result, has_item(item_definition_updated), "Transaction: item wasn't updated")

    def test_delete_item(self, db, namespace, index, item):
        # Given("Create namespace with item")
        # When ("Delete item")
        delete_item_transaction(db, namespace, item)
        # Then ("Check that item is deleted")
        select_result = get_ns_items(db, namespace)
        assert_that(select_result, empty(), "Transaction: item wasn't deleted")

    def test_rollback_transaction(self, db, namespace, index):
        # Given("Create namespace with index")
        # When ("Insert items into namespace")
        transaction = db.tx.begin(namespace)
        number_items = 5
        for i in range(number_items):
            transaction.insert_item({"id": i, "field": "value"})
        # Then ("Rollback transaction")
        transaction.rollback(timeout=timedelta(milliseconds=1000))
        # When ("Get namespace information")
        select_result = get_ns_items(db, namespace)
        # Then ("Check that list of items in namespace is empty")
        assert_that(select_result, empty(), "Transaction: item list is not empty")

    def test_commit_tx_timeout_small(self, db, namespace, index):
        """ Check that timeout is only for tx begin/commit methods, and not for the whole tx """
        # Given("Create namespace with index")
        # When ("Begin tx with small timeout, insert item")
        transaction = db.tx.begin(namespace, timeout=timedelta(milliseconds=20))
        time.sleep(0.1)
        transaction.insert_item(item_definition)
        # When ("Commit tx with small timeout")
        transaction.commit(timeout=timedelta(milliseconds=20))
        # Then ("Check that item was added")
        select_result = get_ns_items(db, namespace)
        assert_that(select_result, has_length(1), "Transaction: item wasn't created")
        assert_that(select_result, has_item(item_definition), "Transaction: item wasn't created")

    def test_transaction_query_delete(self, db, namespace, index, items):
        # Given("Create namespace with items")
        # When ("Delete items with transaction")
        transaction = db.tx.begin(namespace)
        query = db.query.new(namespace)
        query.where('id', CondType.CondGe, 0)
        transaction.delete_query(query)
        transaction.commit(timeout=timedelta(milliseconds=1000))
        # Then ("Check that items is deleted")
        select_result = get_ns_items(db, namespace)
        assert_that(select_result, empty(), "Transaction: item wasn't deleted")

    def test_transaction_query_update(self, db, namespace, index, item):
        # Given("Create namespace with item")
        # When ("Update item")
        item_definition_updated = {'id': 100, 'val': "new_value"}
        transaction = db.tx.begin(namespace)
        query = db.query.new(namespace)
        query.where('id', CondType.CondEq, 100).set('val', ['new_value'])
        transaction.update_query(query)
        transaction.commit(timeout=timedelta(milliseconds=1000))
        # Then ("Check that item is updated")
        select_result = get_ns_items(db, namespace)
        assert_that(select_result, has_length(1), "Transaction: item wasn't updated")
        assert_that(select_result, has_item(item_definition_updated), "Transaction: item wasn't updated")



import random

from datetime import timedelta
from typing import Final, List

from pyreindexer import RxConnector
from pyreindexer.index_search_params import IndexSearchParamHnsw
from pyreindexer.exceptions import ApiError
from pyreindexer.query import CondType


def create_index_example(db, namespace):
    index_definition = {
        'name': 'id',
        'json_paths': ['id'],
        'field_type': 'int',
        'index_type': 'hash',
        'is_pk': True,
        'is_array': False,
        'is_dense': False,
        'is_sparse': False,
        'is_no_column': False,
        'collate_mode': 'none',
        'sort_order_letters': '',
        'expire_after': 0,
        'config': {},
    }

    try:
        db.index_add(namespace, index_definition)
    except ApiError:
        db.index_drop(namespace, 'id', timedelta(milliseconds = 1000))
        db.index_add(namespace, index_definition)


def update_index_example(db, namespace):
    index_definition_modified = {
        'name': 'id',
        'json_paths': ['id'],
        'field_type': 'int64',
        'index_type': 'hash',
        'is_pk': True,
        'is_array': False,
        'is_dense': True,
        'is_sparse': False,
        'is_no_column': False,
        'collate_mode': 'none',
        'sort_order_letters': '',
        'expire_after': 0,
        'config': {},
    }
    db.index_update(namespace, index_definition_modified)


def create_items_example(db, namespace):
    items_count = 10

    for i in range(0, items_count):
        item = {'id': 1, 'name': 'item_' + str(i % 2), 'value': 'check'}
        db.item_upsert(namespace, item, ["id=serial()"])


def select_item_query_example(db, namespace):
    item_name_for_lookup = 'item_0'

    return db.exec_sql(f"SELECT * FROM {namespace} WHERE name='{item_name_for_lookup}'", timedelta(milliseconds = 1000))

def select_all_item_query_example(db, namespace):
    return db.exec_sql(f'SELECT * FROM {namespace}', timedelta(milliseconds = 1000))

def print_all_records_from_namespace(db, namespace, message):
    selected_items_tr = select_all_item_query_example(db, namespace)

    res_count = selected_items_tr.count()
    print(message, res_count)

    for item in selected_items_tr:
        print(f'item: {item}')

def transaction_example(db, namespace, items_in_base):
    # start transaction
    transaction = db.new_transaction(namespace)

    items_count = len(items_in_base)

    # delete first few items
    for i in range(int(items_count / 2)):
        transaction.delete(items_in_base[i])

    # update last one item, overwrite field 'value'
    item = items_in_base[items_count - 1]
    item['value'] = 'transaction was here'
    transaction.update(item)

    # stop transaction and commit changes to namespace
    count = transaction.commit_with_count(timedelta(milliseconds = 1000))
    print(f'Transaction updated count: {count}')

    print_all_records_from_namespace(db, namespace, 'Transaction results count: ')

def query_example(db, namespace):
    # query all items
    any_items = (db.new_query(namespace)
                 .where('value', CondType.CondAny)
                 .sort('id')
                 .execute())
    print(f'Query results count (Any): {any_items.count()}')
    for item in any_items:
        print(f'item: {item}')

    # query some items
    selected_items = (db.new_query(namespace)
                      .where('value', CondType.CondEq, 'check')
                      .sort('id')
                      .limit(4)
                      .execute(timedelta(milliseconds = 1000)))
    print(f'Query results count (limited): {selected_items.count()}')
    for item in selected_items:
        print(f'item: {item}')

    # delete some items
    del_count = (db.new_query(namespace)
                 .where('name', CondType.CondEq, 'item_1')
                 .delete(timedelta(milliseconds = 1000)))
    print(f'Deleted count: {del_count}')

    # query all actual items
    any_items = (db.new_query(namespace)
                 .where('value', CondType.CondAny)
                 .must_execute())
    print(f'Query results count (Any after delete): {any_items.count()}')
    for item in any_items:
        print(f'item: {item}')

def modify_query_transaction(db, namespace):
    # start transaction
    transaction = db.new_transaction(namespace)

    # create an update query and set it for the transaction
    query_upd = db.new_query(namespace).where("id", CondType.CondLe, 5).set("name", ["update_with_query_tx"])
    transaction.update_query(query_upd)

    # create a delete query and set it for the transaction
    query_del = db.new_query(namespace).where("id", CondType.CondGe, 6)
    transaction.delete_query(query_del)

    # stop transaction and commit changes to namespace
    transaction.commit(timedelta(milliseconds = 1000))

    print_all_records_from_namespace(db, namespace, 'Transaction with Query results count: ')

def random_vector(dimension: int) -> List[float]:
    return [random.uniform(-10.0, 10.0) for _ in range(dimension)]

def float_vector_hnsw_example(db):
    namespace = 'knn_hnsw'
    db.namespace_open(namespace)

    # create index
    fv_index_name = 'hnsw_idx'
    dimension: Final[int] = 4
    index_definitions = [{'name': 'id',
                          'json_paths': ['id'],
                          'field_type': 'int',
                          'index_type': 'hash',
                          'is_pk': True,
                          'is_array': False,
                          'is_dense': False,
                          'is_sparse': False,
                          'is_no_column': False,
                          'collate_mode': 'none',
                          'sort_order_letters': '',
                          'expire_after': 0,
                          'config': {}},
                         {"name": fv_index_name,
                          "json_paths": [fv_index_name],
                          "field_type": "float_vector",
                          "index_type": "hnsw",
                          "config": {
                              "dimension": dimension,
                              "metric": "inner_product",
                              "start_size": 100,
                              "m": 16,
                              "ef_construction": 200,
                              "multithreading": 1}}]
    for index in index_definitions:
        db.index_add(namespace, index)

    # generate items
    transaction = db.new_transaction(namespace)
    for i in range(100):
        transaction.insert({"id": 0, fv_index_name: random_vector(dimension)}, ["id=serial()"])
    transaction.commit(timedelta(seconds = 3))

    # do query
    param = IndexSearchParamHnsw(k=20, ef=30)
    query_result = (db.new_query(namespace)
                        .where_knn(fv_index_name, random_vector(dimension), param)
                        .select_fields("vectors()")
                        .with_rank()
                        .sort(index="rank()", desc=True)
                        .must_execute(timedelta(seconds = 1)))

    # result
    print("HNSW where_knn: ", query_result.count())
    for item in query_result:
        print('item vec: ', item, end='\n')

    # drop index
    db.index_drop(namespace, fv_index_name, timedelta(milliseconds = 300))

def float_vector_brute_force_sql_example(db):
    namespace = 'knn_bf'
    db.namespace_open(namespace)

    # create index
    fv_index_name = 'bf_idx'
    dimension: Final[int] = 4
    index_definitions = [{'name': 'id',
                          'json_paths': ['id'],
                          'field_type': 'int',
                          'index_type': 'hash',
                          'is_pk': True,
                          'is_array': False,
                          'is_dense': False,
                          'is_sparse': False,
                          'is_no_column': False,
                          'collate_mode': 'none',
                          'sort_order_letters': '',
                          'expire_after': 0,
                          'config': {}},
                         {"name": fv_index_name,
                          "json_paths": [fv_index_name],
                          "field_type": "float_vector",
                          "index_type": "vec_bf",
                          "config": {
                              "dimension": 3,
                              "metric": "inner_product",
                              "start_size": 10000}}]
    for index in index_definitions:
        db.index_add(namespace, index)

    # update index
    index_definition_modified = {"name": fv_index_name,
                                 "json_paths": [fv_index_name],
                                 "field_type": "float_vector",
                                 "index_type": "vec_bf",
                                 "config": {
                                     "dimension": dimension,
                                     "metric": "l2",
                                     "start_size": 1000}}
    db.index_update(namespace, index_definition_modified)

    # generate items
    transaction = db.new_transaction(namespace)
    for i in range(100):
        transaction.insert({"id": i, fv_index_name: random_vector(dimension)})
    transaction.commit(timedelta(seconds = 3))

    # execute SQL query SELECT KNN
    value = random_vector(dimension)
    k: Final[int] = 27
    query = f'SELECT *, vectors() FROM {namespace} WHERE KNN({fv_index_name}, {value}, k={k})'
    query_result = db.exec_sql(query, timedelta(seconds = 1))
    print("Select where KNN: ", query_result.count())
    for item in query_result:
        print('item vec: ', item, end='\n')

    # drop index
    db.index_drop(namespace, fv_index_name, timedelta(milliseconds = 300))


def rx_example():

    db = RxConnector(f'builtin:///tmp/pyrx', max_replication_updates_size = 10 * 1024 * 1024)
    #db = RxConnector('cproto://127.0.0.1:6534/pyrx', enable_compression = True, fetch_amount = 500)

    namespace = 'test_table'
    db.namespace_open(namespace)

    create_index_example(db, namespace)
    update_index_example(db, namespace)

    create_items_example(db, namespace)
    print_all_records_from_namespace(db, namespace, 'All items: ')

    selected_items = select_item_query_example(db, namespace)

    res_count = selected_items.count()
    print(f'Results count: {res_count}')

    # disposable QueryResults iterator
    items_copy = []
    for item in selected_items:
        items_copy.append(item)
        print(f'item: {item}')

    # won't be iterated again
    for item in selected_items:
        print(f'item: {item}')

    transaction_example(db, namespace, items_copy)

    query_example(db, namespace)

    modify_query_transaction(db, namespace)

    float_vector_hnsw_example(db)
    float_vector_brute_force_sql_example(db)

    db.close()


if __name__ == "__main__":
    rx_example()



from hamcrest import *

from tests.helpers.matchers import close_to_dict


def check_response_has_close_to_ns_items(res_items, ns_items, delta=0.01):
    assert_that(res_items, not_(empty()))
    for res_item in res_items:
        assert_that(ns_items, has_item(close_to_dict(res_item, delta)))


def check_response_has_only_close_to_items(res_items, items):
    assert_that(res_items, has_length(len(items)))
    for item in res_items:
        assert_that(items, has_item(close_to_dict(item)))
        found_item = [i for i in items if i["id"] == item["id"]][0]
        items.remove(found_item)
    assert_that(items, empty())



import copy
from typing import Final

import pytest
from hamcrest import *

from pyreindexer.exceptions import ApiError, QueryError
from pyreindexer.index_search_params import IndexSearchParamBruteForce, IndexSearchParamHnsw, IndexSearchParamIvf
from tests.helpers.base_helper import random_vector
from tests.helpers.check_helper import check_response_has_close_to_ns_items
from tests.test_data.constants import vector_index_bf, vector_index_hnsw, vector_index_ivf


class TestQueryWhereKNNString:

    def test_query_where_knn_string_bf(self, db, namespace, index):
        # Given ("Create float vector index")
        dimension: Final[int] = 5
        index = copy.deepcopy(vector_index_bf)
        index["config"]["dimension"] = dimension
        index["config"]["embedding"] = {
            "query_embedder": {"URL": "http://127.0.0.1:8000"}
        }
        db.index.create(namespace, index)
        # Given("Insert items")
        items = [{"id": i, "vec": random_vector(dimension)} for i in range(10)]
        for item in items:
            db.item.insert("new_ns", item)
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Execute query")
        k: Final[int] = 5
        param = IndexSearchParamBruteForce(k=k)
        query_result = list(
            query.where_knn_string("vec", "word", param)
            .select_fields("vectors()")
            .execute())
        # Then ("Check knn select result")
        check_response_has_close_to_ns_items(query_result, items)
        assert_that(query_result, has_length(k))

    def test_query_where_knn_string_hnsw(self, db, namespace, index):
        # Given ("Create float vector index")
        dimension: Final[int] = 5
        index = copy.deepcopy(vector_index_hnsw)
        index["config"]["dimension"] = dimension
        index["config"]["embedding"] = {
            "query_embedder": {"URL": "http://127.0.0.1:8000"}
        }
        db.index.create(namespace, index)
        # Given("Insert items")
        items = [{"id": i, "vec": random_vector(dimension)} for i in range(10)]
        for item in items:
            db.item.insert("new_ns", item)
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Execute query")
        param = IndexSearchParamHnsw(k=5, ef=5)
        query_result = list(
            query.where_knn_string("vec", "same same", param)
            .select_fields("vectors()")
            .execute())
        # Then ("Check knn select result")
        check_response_has_close_to_ns_items(query_result, items)

    def test_query_where_knn_string_ivf(self, db, namespace, index):
        # Given ("Create float vector index")
        dimension: Final[int] = 5
        index = copy.deepcopy(vector_index_ivf)
        index["config"]["dimension"] = dimension
        index["config"]["embedding"] = {
            "query_embedder": {"URL": "http://127.0.0.1:8000"}
        }
        db.index.create(namespace, index)
        # Given("Insert items")
        items = [{"id": i, "vec": random_vector(dimension)} for i in range(10)]
        for item in items:
            db.item.insert("new_ns", item)
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Execute query")
        param = IndexSearchParamIvf(k=5, nprobe=2)
        query_result = list(
            query.where_knn_string("vec", "a b c", param)
            .select_fields("vectors()")
            .execute())
        # Then ("Check knn select result")
        check_response_has_close_to_ns_items(query_result, items)

    @pytest.mark.parametrize("value, err_type, err_msg", [
        (None, TypeError, "object of type 'NoneType' has no len()"),
        (1, TypeError, "object of type 'int' has no len()"),
        (["abc"], TypeError, "must be str, not list"),
        ({"a": "b"}, TypeError, "must be str, not dict"),
        ("", QueryError, "A required parameter is not specified. `value` can't be None or empty"),
    ])
    def test_query_where_knn_string_invalid_value(self, db, namespace, index, value, err_type, err_msg):
        # Given ("Create float vector index")
        dimension: Final[int] = 5
        index = copy.deepcopy(vector_index_hnsw)
        index["config"]["dimension"] = dimension
        index["config"]["embedding"] = {
            "query_embedder": {"URL": "http://127.0.0.1:8000"}
        }
        db.index.create(namespace, index)
        # Given("Insert items")
        items = [{"id": i, "vec": random_vector(dimension)} for i in range(10)]
        for item in items:
            db.item.insert("new_ns", item)
        # Given ("Create new query")
        param = IndexSearchParamHnsw(k=5, ef=5)
        query = db.query.new(namespace).select_fields("vectors()")
        # When ("Try to add where_knn_string to the query")
        assert_that(calling(query.where_knn_string).with_args("vec", value, param),
                    raises(err_type, pattern=err_msg))

    def test_query_where_knn_string_no_query_embedder_config(self, db, namespace, index):
        # Given ("Create float vector index")
        dimension: Final[int] = 5
        index = copy.deepcopy(vector_index_hnsw)
        index["config"]["dimension"] = dimension
        db.index.create(namespace, index)
        # Given("Insert items")
        items = [{"id": i, "vec": random_vector(dimension)} for i in range(10)]
        for item in items:
            db.item.insert("new_ns", item)
        # Given ("Create new query")
        param = IndexSearchParamHnsw(k=5, ef=5)
        query = db.query.new(namespace).where_knn_string("vec", "word", param).select_fields("vectors()")
        # When ("Try to execute query without query_embedder")
        param = IndexSearchParamHnsw(k=5, ef=5)
        query = query.where_knn_string("vec", "word", param).select_fields("vectors()")
        err_msg = ("Failed to get embedding for 'vec'. Problem with client: Unexpected problem with client|"
                   "Trying to find knn by string. No Embedder configured for index 'vec'")
        assert_that(calling(query.execute).with_args(), raises(ApiError, pattern=err_msg))



import shutil
from datetime import timedelta

import pytest
from hamcrest import *

from pyreindexer.exceptions import ApiError
from tests.helpers.api import ConnectorApi
from tests.helpers.base_helper import prepare_ns_with_items


CONNECTIONS = []


@pytest.fixture
def db(request):
    if request.config.getoption('--mode') != "cproto":
        pytest.skip("tests only for cproto mode")

    def _inner(**kwargs):
        db = ConnectorApi("cproto://127.0.0.1:6534/test_db", **kwargs)
        CONNECTIONS.append(db)
        return db

    yield _inner


@pytest.fixture(autouse=True)
def clear_storage(db):
    yield
    for db in CONNECTIONS:
        for ns in db.namespace.enumerate():
            if not ns["name"].startswith("#"):
                db.namespace.drop(ns["name"])
        db.close()
    CONNECTIONS.clear()
    shutil.rmtree("tmp/", ignore_errors=True)


class TestCprotoOptions:
    def test_cproto_options(self, db):
        db = db(fetch_amount=10, reconnect_attempts=1, net_timeout=timedelta(milliseconds=1000),
                enable_compression=True, start_special_thread=True, client_name="qa_tests", sync_rxcoro_count=1)
        prepare_ns_with_items(db)

    def test_cproto_options_zero_values(self, db):
        db = db(reconnect_attempts=0, net_timeout=timedelta(milliseconds=0), sync_rxcoro_count=1)
        ns_name = "test_zero_val"
        items = prepare_ns_with_items(db, ns_name)
        query = f"SELECT * FROM {ns_name}"
        result = list(db.query.sql(query))
        assert_that(result, equal_to(items))

    def test_cproto_options_zero_sync_rxcoro_count(self, db):
        err_msg = ("The number of synchronization coroutines (SyncRxCoroCount in conf) must be"
                   " in the range [1..10'000]. But was: 0")
        assert_that(calling(db).with_args(sync_rxcoro_count=0), raises(ApiError, matching=has_string(err_msg)))

    def test_cproto_options_fetch_amount_zero_value(self, db):
        assert_that(calling(db).with_args(fetch_amount=0),
                    raises(ValueError, pattern="'fetch_amount' must be greater than zero"))

    def test_cproto_options_negative_values(self, db):
        db = db(reconnect_attempts=-1, net_timeout=timedelta(milliseconds=-1), sync_rxcoro_count=1)
        ns_name = "test_negative_val"
        items = prepare_ns_with_items(db, ns_name)
        query = f"SELECT * FROM {ns_name}"
        result = list(db.query.sql(query))
        assert_that(result, equal_to(items))

    def test_cproto_options_negative_sync_rxcoro_count(self, db):
        err_msg = ("The number of synchronization coroutines (SyncRxCoroCount in conf) must be in"
                   " the range [1..10'000]. But was: 4294967295")
        assert_that(calling(db).with_args(sync_rxcoro_count=-1), raises(ApiError, matching=has_string(err_msg)))

    def test_cproto_fetch_amount(self, db):
        """ Check that fetch amount doesn't affect the amount of total items returned """
        db = db(fetch_amount=1)
        ns_name = "test_fetch"
        items = prepare_ns_with_items(db, ns_name)
        query = f"SELECT * FROM {ns_name}"
        result = list(db.query.sql(query))
        assert_that(result, equal_to(items))

class TestCprotoFunc:
    def test_cproto_query_fail(self, db):
        # NOTE: don't do this, it's wrong - just for testing
        db = db(fetch_amount=1)
        ns_name = "test_query_fail_fetch"
        items = prepare_ns_with_items(db, ns_name)
        ct = db.new_query(ns_name).select_fields('id').execute().count()
        assert_that(ct, not_(equal_to(0)))



USER_ROLES = ["data_read", "data_write", "db_admin", "owner"]

users_yml = '''
data_read:
    hash: dataread
    roles:
        *: data_read

data_write:
    hash: datawrite
    roles:
        *: data_write

db_admin:
    hash: dbadmin
    roles:
        *: db_admin

owner:
    hash: owner
    roles:
        *: owner
        
invalid_pass:
    hash: invalid_pass
    roles:
        *: owner
'''



from pyreindexer.exceptions import ApiError, TransactionError


class RaiserRx:
    """Contains methods for checking some typical API bad events and raise if there is a necessity

    """
    err_code: int
    err_msg: str
    rx: int

    @staticmethod
    def raise_if_error(func):
        def wrapper(self, *args, **kwargs):
            self.raise_on_not_init()
            res = func(self, *args, **kwargs)
            self.raise_on_error()
            return res

        return wrapper

    def raise_on_error(self):
        """Checks if there is an error code and raises with an error message

        #### Raises:
            ApiError: Raises with an error message of API return on non-zero error code

        """

        if self.err_code:
            raise ApiError(self.err_msg)

    def raise_on_not_init(self):
        """Checks if there is an error code and raises with an error message

        #### Raises:
            ConnectionError: Raises with an error message when Reindexer instance is not initialized yet

        """

        if self.rx <= 0:
            raise ConnectionError("Connection is not initialized: reindexer C-binding object does not exist")


class RaiserQuery(RaiserRx):
    """Contains methods for checking some typical API bad events and raise if there is a necessity

    """
    rx: None

    def raise_on_not_init(self):
        """Checks if there is an error code and raises with an error message

        #### Raises:
            ConnectionError: Raises with an error message when Reindexer instance is not initialized yet

        """

        if self.rx.rx <= 0:
            raise ConnectionError("Connection is not initialized: reindexer C-binding object does not exist")


class RaiserTx(RaiserQuery):
    """Contains methods for checking some typical API bad events and raise if there is a necessity

    """
    transaction_wrapper_ptr: int

    @staticmethod
    def raise_if_error(func):
        def wrapper(self, *args, **kwargs):
            self.raise_on_not_init()
            self.raise_on_is_over()
            res = func(self, *args, **kwargs)
            self.raise_on_error()
            return res

        return wrapper

    def raise_on_is_over(self):
        """Checks if there is an error code and raises with an error message

        #### Raises:
            ApiError: Raises with an error message of API return on non-zero error code

        """

        if self.transaction_wrapper_ptr <= 0:
            raise TransactionError("Transaction is over")



import math
import random
from typing import List

from tests.test_data.constants import index_definition


def get_ns_items(db, ns_name):
    """ Get all items via sql query
    """
    return list(db.query.sql(f"SELECT * FROM {ns_name}"))

def get_ns_vect_items(db, ns_name):
    """ Get all items via sql query
    """
    return list(db.query.sql(f"SELECT *, vectors() FROM {ns_name}"))

def get_ns_description(db, ns_name):
    """ Get information about namespace in database
    """
    namespaces_list = db.namespace.enumerate()
    ns_entry = [ns for ns in namespaces_list if ns["name"] == ns_name]
    return ns_entry


def prepare_ns_with_items(db, ns_name="new_ns", items_num=5) -> list:
    """ Create ns, index and items
    """
    db.namespace.open(ns_name)
    db.index.create(ns_name, index_definition)
    items = [{"id": i, "val": f"testval{i}"} for i in range(items_num)]
    for item in items:
        db.item.insert(ns_name, item)
    return items


def calculate_distance(point1, point2):
    return math.sqrt((point1[0] - point2[0]) ** 2 + (point1[1] - point2[1]) ** 2)


def random_vector(dimension: int) -> List[float]:
    return [random.uniform(-10.0, 10.0) for _ in range(dimension)]
